# Plugin for TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2008-2014 Hideyo Imazu, hideyo.imazu[at]morganstanley.com and
# TWiki Contributors.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html
#
# As per the GPL, removal of this notice is prohibited.
#
# =========================
#
# This is TWiki's Data Table Plugin

use strict;

=pod

---+ package DataTablePlugin

=cut

package TWiki::Plugins::DataTablePlugin;

use TWiki::Plugins::DataTablePlugin::Core;

our $VERSION = '$Rev: 29965 (2016-02-16) $';
our $RELEASE = '2014-03-12';
our $SHORTDESCRIPTION = 'Look up a table on a TWiki topic and display the result in a specified format a la SEARCH';
our $NO_PREFS_IN_TOPIC = 1;
our $pluginName = 'DataTablePlugin';
our $debug = 0;
our $initialized = 0;

=pod

---++ initPlugin($topic, $web, $user, $installWeb) -> $boolean
   * =$topic= - the name of the topic in the current CGI query
   * =$web= - the name of the web in the current CGI query
   * =$user= - the login name of the user
   * =$installWeb= - the name of the web the plugin is installed in

=cut

sub initPlugin {
    my( $topic, $web, $user, $installWeb ) = @_;

    # check for Plugins.pm versions
    if( $TWiki::Plugins::VERSION < 1.026 ) {
        TWiki::Func::writeWarning(
	    "Version mismatch between $pluginName and Plugins.pm" );
        return 0;
    }

    # register the TAG functions to handle %TAG{...}%
    # This will be called whenever %TAG% or %TAG{...}% is
    # seen in the topic text.
    TWiki::Func::registerTagHandler( 'DATATBL', \&_DATATBL );

    $initialized = 0; # flag that real initialization is not done

    # Plugin correctly initialized
    return 1;
}

sub expandSub {
    my ($colNameCan, $row, $tbl, $assumedCol, $trFrom, $trTo, $trOpt) = @_;
    my $colName = substr($colNameCan, 1);
    $colName =~ s/\(\)$//;
    if ( $row->{$colName} ) {
	my $retVal = join(', ', @{$row->{$colName}});
	if ( $trFrom ne '' ) {
	    eval "\$retVal =~ tr/$trFrom/$trTo/$trOpt";
	}
	return $retVal;
    }
    elsif ( $tbl->{cols}{$colName} || $assumedCol->{$colName} ) {
	return '';
    }
    else {
	return $colNameCan;
    }
}

sub expand {
    my ($format, $tbl, $rowNum, $for, $assumed, $trF, $trT, $trO, $resultSeq) =
        @_;
    my $rowNum1 = $rowNum + 1;
    my $row = $tbl->{rows}[$rowNum];
    my @iter;
    if ( $for ) {
	if ( $row->{$for} ) {
	    @iter = @{$row->{$for}};
	}
    }
    else {
	@iter = ('');
    }
    my @expandeds;
    for my $i ( @iter ) {
	my $expanded = $format;
	$expanded =~
s/(\$$COLNAME_RE(\(\))?)/expandSub($1, $row, $tbl, $assumed, $trF, $trT, $trO)/geo;
	$expanded =~ s/\$i\(\)/$i/g;
	$expanded =~ s/\$i\b/$i/g;
	$expanded =~ s/\$rownum\(\)/$rowNum/g;
	$expanded =~ s/\$rownum\b/$rowNum/g;
	$expanded =~ s/\$rownum1\(\)/$rowNum1/g;
	$expanded =~ s/\$rownum1\b/$rowNum1/g;
	$expanded =~ s/\$resultseq\(\)/$resultSeq/g;
	$expanded =~ s/\$resultseq\b/$resultSeq/g;
	push(@expandeds, TWiki::Func::decodeFormatTokens($expanded));
    }
    return @expandeds;
}

sub init {
    # my ($session, $params, $theTopic, $theWeb) = @_;
    #     $_[0]     $_[1]    $_[2]      $_[3]
    $debug = TWiki::Func::isTrue(
	TWiki::Func::getPreferencesValue("DATATBL_DEBUG", $_[0]->{webName}));
    $initialized = 1;
}

sub procTrStr {
    my $str = shift;
    return '' unless ( defined($str) );
    $str = TWiki::Func::decodeFormatTokens($str);
    $str =~ s:/::g;
    $str =~ s:\n::g;
    $str =~ s/\\+$//;
    return $str;
}

use Data::Dumper;

sub _DATATBL {
    init(@_) unless ( $initialized );
    my ($session, $params, $theTopic, $theWeb) = @_;
    $theWeb ||= $session->{webName};
	# When a new topic is created, $theWeb is null
    TWiki::Func::writeDebug('_DATATBL called ' . Dumper($params))
	if ( $debug );
    $session->{DataTblCore} ||=
	TWiki::Plugins::DataTablePlugin::Core->new($session, $theWeb, $debug);
    my $tblName = $params->{_DEFAULT} || $params->{tbl} || return '';
    my $ifNoMatch = $params->{ifnomatch} || '';
    my $ignoreNull = TWiki::Func::isTrue($params->{ignorenull});
    my $limit = $params->{limit} || '';
    if ( $limit =~ /^(\d+)$/ ) {
        $limit = $1;
    }
    else {
        $limit = 0;
    }
    my $transposed = TWiki::Func::isTrue($params->{transposed});
    my $split = $params->{split};
    $split = '' unless ( defined($split) );
    my $aCols = $params->{assumedcols};
    $aCols = '' unless ( defined($aCols) );
    my %assumedCol;
    %assumedCol = map { $_ => 1 } split(/, */, $aCols) if ( $aCols ne '' );
    my $trFrom = procTrStr($params->{trfrom});
    my $trTo = procTrStr($params->{trto});
    my $trOptStr = $params->{tropt} || '';
    my $trOpt = '';
    if ( $trOptStr ) {
	$trOpt .= 'c' if $trOptStr =~ /c/i;
	$trOpt .= 'd' if $trOptStr =~ /d/i;
	$trOpt .= 's' if $trOptStr =~ /s/i;
    }
    my $showNRows = TWiki::Func::isTrue($params->{shownrows});
    my ($by, $value);
    my @bys = sort grep { /^[A-Z]/ } keys %$params;
    if ( @bys ) {
	$by = $bys[0];
	$value = $params->{$by};
    }
    else {
	$by = $params->{by} || '';
	$value = $params->{value};
    }
    my $rows = $params->{rows};
    $rows = '' unless defined($rows);
    my $for = $params->{for} || ''; 
    my $format = $params->{format} || '$Name';
    # separator
    my %separators;
    for my $i ( keys %$params ) {
	if ( $i =~ /^(separator(\d*))$/ ) {
	    my ($pname, $num) = ($1, $2);
	    $num ||= 1;
	    $separators{$num} = TWiki::expandStandardEscapes($params->{$pname});
	}
    }
    my @sepIndexes = sort {$a <=> $b} keys %separators;
    my $evalCell = TWiki::Func::isTrue($params->{evalcell});
    my $ignoreCase = TWiki::Func::isTrue($params->{ignorecase});
    my $tbl = $session->{DataTblCore}->get($tblName, $transposed, $split,
					   $evalCell) or
	return $ifNoMatch;
    my $rowNumMax = $#{$tbl->{rows}};
    my @rowNums;
    my @allRowNums = 0..$rowNumMax;
    if ( $by && defined($value) ) {
	my $inverse = ($value =~ s/^!\s*//) ? 1 : 0;
	my @values = split(/\s*,\s*/, $value);
	my %rowNum;
	for my $i ( @values ) {
	    next if ( $i eq '' );
	    my $idx = $i eq "''"  ? $tbl->{indexes}{$by}{"\0"} :
		      $ignoreCase ? $tbl->{ciindexes}{$by}{lc $i} :
				    $tbl->{indexes}{$by}{$i};
	    if ( $i eq "''" ) {
		if ( $idx ) {
		    my @a = grep { !$idx->{$_} } @allRowNums;
		    @rowNum{@a} = (1) x @a;
		}
		else {
		    %rowNum = map { $_ => 1 } @allRowNums;
		    last;
		}
	    }
	    else {
		if ( $idx ) {
		    @rowNum{keys %$idx} = values %$idx;
		}
	    }
	    # TWiki::Func::writeDebug("_DATATBL \$i = $i, \%rowNum = " . Dumper(\%rowNum));
	}
	if ( $inverse ) {
	    %rowNum = map { $_ => 1 } grep { !$rowNum{$_} } @allRowNums;
	}
	if ( (keys %rowNum) == 0 ) {
	    TWiki::Func::writeDebug("_DATATBL no match, returning $ifNoMatch")
		if ( $debug );
	    return $ifNoMatch;
	}
	@rowNums = sort {$a <=> $b} keys %rowNum;
    }
    elsif ( $rows ne '' ) {
	my %rowsHash = map { $_ => 1 } split(/\s*,\s*/, $rows);
	@rowNums = grep { $rowsHash{$_} } @allRowNums;
    }
    else {
	@rowNums = @allRowNums;
    }
    return scalar(@rowNums) if ( $showNRows );
    my $sortCol;
    if ( $sortCol = $params->{sort} || $params->{sorta} ) {
        if ( $tbl->{cols}{$sortCol} ) {
            my %sortKey;
            for my $i ( @rowNums ) {
                my $key = join(', ', @{$tbl->{rows}[$i]{$sortCol}});
                unless ( $params->{sort} ) {
                    $key = lc $key;
                }
                $sortKey{$i} = $key;
            }
            @rowNums = sort { $sortKey{$a} cmp $sortKey{$b} } @rowNums;
        }
    }
    elsif ( $sortCol = $params->{sortn} ) {
        if ( $tbl->{cols}{$sortCol} ) {
            my %sortKey;
            for my $i ( @rowNums ) {
                $sortKey{$i} = $tbl->{rows}[$i]{$sortCol}[0];
            }
            @rowNums = sort { $sortKey{$a} <=> $sortKey{$b} } @rowNums;
        }
    }
    if ( TWiki::Func::isTrue($params->{reverse}) ) {
        @rowNums = reverse @rowNums;
    }
    if ( $limit ) {
	splice(@rowNums, $limit);
    }
    my @expandeds;
    my $resultSeq = 1;
    for my $i ( @rowNums ) {
	push(@expandeds,
	     expand($format, $tbl, $i, $for, \%assumedCol,
		    $trFrom, $trTo, $trOpt, $resultSeq++));
    }
    if ( $ignoreNull ) {
        @expandeds = grep { defined $_ && $_ ne '' } @expandeds;
    }
    if ( @expandeds == 0 ) {
	TWiki::Func::writeDebug("_DATATBL no match, returning $ifNoMatch")
	    if ( $debug );
	return $ifNoMatch;
    }
    if ( @sepIndexes == 0 ) {
	my $retVal = join('', @expandeds);
	TWiki::Func::writeDebug("_DATATBL returning $retVal")
	    if ( $debug );
	return $retVal;
    }
    if ( @sepIndexes == 1 && $sepIndexes[0] == 1 ) {
	my $retVal = join($separators{1}, @expandeds);
	TWiki::Func::writeDebug("_DATATBL returning $retVal")
	    if ( $debug );
	return $retVal;
    }
    my @sepToUse = '' x @expandeds;
    for my $i ( @sepIndexes ) {
	my $sep = $separators{$i};
	for ( my $j = $i ; $j < @expandeds ; $j += $i ) {
	    $sepToUse[$j] = $sep;
	}
    }
    my $retVal = $expandeds[0];
    for my $i ( 1..$#expandeds ) {
	$retVal .= $sepToUse[$i] . $expandeds[$i];
    }
    TWiki::Func::writeDebug("_DATATBL returning $retVal")
	if ( $debug );
    return $retVal;
}

1;
