# Plugin for TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2008-2014 Hideyo Imazu, hideyo.imazu[at]morganstanley.com and
# TWiki Contributors.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html
#
# As per the GPL, removal of this notice is prohibited.
#
# =========================
#
# This is part of TWiki's Data Table Plugin.
#
# The code below is kept out of the main plugin module for
# performance reasons, so it doesn't get compiled until it
# is actually used.
# Actually, the above goal is not achieved as of now.

use strict;

package TWiki::Plugins::DataTablePlugin::Core;

use Data::Dumper;

use base 'Exporter';
our @EXPORT = qw($COLNAME_RE);

our $COLNAME_RE = qr/[A-Za-z]\w*/;
our $COLNAME_IN_TBL_RE = qr/[A-Za-z][^*]*/;

sub new {
    my ($class, $session, $webName, $debug) = @_;
    my $self = bless {}, $class;
    $self->{session} = $session;
    $self->{webName} = $webName;
    $self->{debug} = $debug;
    $self->{recursionDepth} = 0;
    $self->{tbl} = {};
    return $self;
}

sub get {
    my ($self, $tblName, $transposed, $split, $evalCell) = @_;
    my $debug = $self->{debug};
    my (@tables, @tableNames, $canonTblName);
    if ( $tblName =~ /\$/ ) {
	$canonTblName = $tblName;
	@tables = ($tblName);
    }
    else {
	for my $i ( split(/\s*,\s*/, $tblName) ) {
	    my @elems = split(m:[./]:, $i);
	    my $w = @elems > 1 ? join('/', @elems[0..$#elems - 1]) :
		$self->{webName};
	    my $t = $elems[-1];
	    push(@tables, [$w, $t]);
	    push(@tableNames, "$w.$t");
	}
	$canonTblName = join(",", @tableNames);
    }
    TWiki::Func::writeDebug(
	"TWiki::Plugins::DataTablePlugin::get($canonTblName) called")
	if ( $debug );
    my $tbl = $self->{tbl}{$canonTblName};
    defined($tbl) and
	return $tbl;
    $self->{tbl}{$canonTblName} = '';
	# in case of an error, you can just return;
    my %theTbl = (rows => [], cols => {}, indexes => {}, ciindexes => {});
    my $rowNum = 0;
    for my $t ( @tables ) {
	my ($web, $topic, $text);
	if ( ref $t )  {
	    ($web, $topic) = @$t;
	    if ( !TWiki::Func::topicExists($web, $topic) ) {
		TWiki::Func::writeDebug("$web.$topic not found")
		    if ( $debug );
		next;
	    }
	    $text = TWiki::Func::readTopicText($web, $topic);
	}
	else {
	    if ( $self->{recursionDepth} > 16 ) {
		TWiki::Func::writeWarn(
"\%DATATBL{\"$t\"...}\% on $self->{session}{topicName}.$self->{theWeb}" .
"caused a too deep recursion.");
		$text = '';
	    }
	    else {
		$self->{recursionDepth}++;
		$t = TWiki::expandStandardEscapes($t);
		$text = TWiki::Func::expandCommonVariables($t,
		    $self->{session}{topicName}, $self->{theWeb});
		TWiki::Func::writeDebug(
				"DataTablePlugin::Core: $t expanded to $text")
		    if ( $debug );
		$self->{recursionDepth}--;
	    }
	}
	# text cleanup
	$text =~ s/\r//g;
	$text =~ s/\\\n//gs;
	if ( $transposed ) {
	    my $inTbl = 0;
	    my $row = {};
	    for my $line ( split(/\n/, $text) ) {
		if ( $line =~ /^\s*\|\s*(.*?)\s*\|\s*$/ &&
		     !($split ne '' && $line =~ /$split/)
		) {
		    my @cols = split(/\s*\|\s*/, $1);
		    $inTbl = 1 unless ( $inTbl );
		    my $colName = '';
		    if ( $cols[0] =~ /($COLNAME_IN_TBL_RE)/o ) {
			$colName = $1;
                        $colName =~ s/\W+$//;
			$colName =~ s/\W+/_/g;
			$theTbl{cols}{$colName} = 1;
		    }
		    my $colVal = $cols[1];
		    $colVal = '' unless ( defined($colVal) );
		    $colVal =~ s/\s*%EDITCELL\{.*\}%$//;
		    $colVal = TWiki::Func::expandCommonVariables(
			$colVal, $topic, $web)
			if ( $evalCell && $colVal =~ /%/ );
		    my @values = grep { $_ ne '' } split(/\s*,\s*/, $colVal);
		    push(@{$row->{$colName} ||= []}, @values);
		    for my $j ( @values ) {
			$theTbl{indexes}{$colName}{$j}{$rowNum} = 1;
			$theTbl{ciindexes}{$colName}{lc $j}{$rowNum} = 1;
		    }
		    if ( @values ) {
			$theTbl{indexes}{$colName}{"\0"}{$rowNum} = 1;
		    }
		}
		else {
		    if ( $inTbl ) {
			push(@{$theTbl{rows}}, $row);
			$row = {};
			$inTbl = 0;
			$rowNum++;
		    }
		}
	    }
	    if ( $inTbl ) {
		push(@{$theTbl{rows}}, $row);
		$rowNum++;
	    }
	}
	else {
	    my $inTbl = 0;
	    my @colNames;
	    for my $line ( split(/\n/, $text) ) {
		if ( $line =~ /^\s*\|\s*(.*?)\s*\|\s*$/ ) {
		    my @cols = split(/\s*\|\s*/, $1);
		    unless ( $inTbl ) {
			$inTbl = 1;
			@colNames = ();
			for my $col ( @cols ) {
			    if ( $col =~ /($COLNAME_IN_TBL_RE)/o ) {
				my $colName = $1;
                                $colName =~ s/\W+$//;
				$colName =~ s/\W+/_/g;
				push(@colNames, $colName);
				$theTbl{cols}{$colName} = 1;
			    }
			}
		    }
		    else {
			my $row = {};
			for my $i ( 0..$#colNames ) {
			    my $colName = $colNames[$i];
			    my $colVal = $cols[$i];
                            $colVal = '' unless ( defined($colVal) );
			    $colVal = TWiki::Func::expandCommonVariables(
				$colVal, $topic, $web)
				if ( $evalCell && $colVal =~ /%/ );
			    my @values = grep { $_ ne '' }
				split(/\s*,\s*/, $colVal);
			    push(@{$row->{$colName} ||= []}, @values);
			    for my $j ( @values ) {
				$theTbl{indexes}{$colName}{$j}{$rowNum} = 1;
				$theTbl{ciindexes}{$colName}{lc $j}{$rowNum}
				    = 1;
			    }
			    if ( @values ) {
				$theTbl{indexes}{$colName}{"\0"}{$rowNum} = 1;
			    }
			}
			push(@{$theTbl{rows}}, $row);
			$rowNum++;
		    }
		}
		else {
		    last if ( $inTbl );
		}
	    }
	}
    }
    TWiki::Func::writeDebug(Dumper(\%theTbl))
	if ( $debug );
    if ( $rowNum == 0 ) {
	return '';
    }
    return $self->{tbl}{$canonTblName} = \%theTbl;
}

1;
